from .Base import Base

import numpy as np
import pystan

class BradleyTerry(Base):
  def __init__(self, alpha, seed=1, num_chains=1, num_samples=2000):
    Base.__init__(self, stan_model='./stan/BradleyTerry.stan', 
                  seed=seed, num_chains=num_chains, num_samples=num_samples)
    self.alpha = alpha

  def get_model_data(self, permus):
    num_permus, num_algorithms = permus.shape

    model_data = {'num_permus': num_permus,
                  'num_algorithms': num_algorithms,
                  'permus': permus,
                  'alpha': self.alpha}
    return model_data

  def get_samples(self, data):
    result = [(ratings, constant) for ratings, constant in zip(data['ratings'], data['constant'])]
    return result

  def calculate_permu_prob(self, permu, params):
    num_algorithms = len(permu)
    ratings, constant = params
    prob = 1

    for i in range(num_algorithms):
      prob *= ratings[permu[i]] ** (num_algorithms - i - 1)
  
    return prob / constant

  




import matplotlib.pyplot as plt
import itertools
import pandas as pd
import numpy as np

class Plot:

  def set_axis_style(self, ax, labels):
    ax.xaxis.set_tick_params(direction='out')
    ax.xaxis.set_ticks_position('bottom')
    ax.set_xticks(np.arange(1, len(labels) + 1))
    ax.set_xticklabels(labels)
    ax.set_xlim(0.25, len(labels) + 0.75)

  def plot_top_ranking_probs(self, model_names, algorithm_names, probs, empirical, axs):
    assert(len(model_names) > 0)
    assert(len(model_names) == len(probs))
    assert(len(algorithm_names) == len(axs))
    num_samples, num_algorithms = probs[0].shape

    for i, ax in enumerate(axs):
      if ax != None:
        df = pd.DataFrame(columns=model_names)

        for model_name, sample in zip(model_names, probs):
          df[model_name] = sample[:,i]
        
        ax.violinplot(df.values, showmeans=True)
        ax.axhline(y=empirical[i], linestyle=':', color='black', label='Empirical')
        ax.set_title(algorithm_names[i])
        self.set_axis_style(ax, model_names)

  def plot_better_than_probs(self, model_names, algorithm_names, probs, empirical, axs):
    assert(len(model_names) > 0)
    assert(len(model_names) == len(probs))
    num_samples, num_algorithms, _ = probs[0].shape

    first_row = True
    for model, prob, row in zip(model_names, probs, axs):
      for i, ax in enumerate(row):
        if ax != None:
          idxs = [x for x in range(0, num_algorithms) if x != i]
          names = [algorithm_names[x] for x in idxs]
          x_values = range(1, len(idxs) + 1)
          empirical_values = [empirical[i, vs] for vs in idxs]
          
          ax.scatter(x_values, empirical_values, marker='o', color='black', s=30, zorder=3)
          ax.violinplot(prob[:, i, idxs], showmeans=True)
          ax.set_ylabel(model)
          self.set_axis_style(ax, names)

          if first_row:
            ax.set_title(algorithm_names[i])
      first_row = False

  def plot_top_k_probs(self, model_names, algorithm_names, probs, empirical, axs):
    assert(len(model_names) > 0)
    assert(len(model_names) == len(probs))
    num_samples, num_algorithms, _ = probs[0].shape

    first_row = True
    for model, prob, row in zip(model_names, probs, axs):
      for i, ax in enumerate(row):
        if ax != None:
          idxs = list(range(num_algorithms))
          names = algorithm_names
          x_values = range(1, len(idxs) + 1)
          empirical_values = [empirical[i, vs] for vs in idxs]
          
          ax.violinplot(prob[:, idxs, i], showmeans=True)
          ax.scatter(x_values, empirical_values, marker='o', color='black', s=30, zorder=3)
          ax.set_ylabel(model)
          self.set_axis_style(ax, names)

          if first_row:
            ax.set_title("Top: " + str(i + 1))
      first_row = False



import os.path
import math
import numpy as np
import itertools
import pystan

class Base:
  def __init__(self, stan_model=None, seed=1, num_chains=1, num_samples=2000):
    self.seed = seed
    self.num_chains = num_chains
    self.num_samples = num_samples
    
    if stan_model != None:
      path = os.path.abspath(os.path.dirname(__file__))
      path = os.path.join(path, stan_model)
      self.code = open(path).read()
      self.posterior = pystan.StanModel(model_code=self.code)

  def sample_posterior(self, permus):
    """ Sample from the posterior distribution.

    Parameters:
    -----------

    permus : ndarray 
      Array of num_permus rows representing a permutation and num_algorithms columns 
      representing  the ranking of each algorithm.

    Returns
    --------

    samples : ndarray
      Array of size (num_posterior_samples, num_model_params) where the first dimension represents
      the number of posterior samples and the second dimension represents the number of parameter
      of the probabilistic model.

    """

    model_data = self.get_model_data(permus)
    fit = self.posterior.sampling(data=model_data, chains=self.num_chains, iter=self.num_samples, seed=self.seed)
    samples = self.get_samples(fit)
    return samples


  def get_model_data(self, permus):
    return {"data": permus}

  def get_samples(self, data):
    return data['ratings']

  def calculate_permu_prob(self, permu, params):
    """ Computes the probability of a given permutation according to a given
    probability model defined by params. Derived classes should provide a concrete
    implementation of this method.

    Parameters:
    -----------

    permu : ndarray
      A one dimensional array representing a permutation.
    params: ndarray
      The parameters of the probability model.

    """
    return 0

  def sample_uniform_permu(self, permu):
    """ Sample an uniform random permutation and store the result in permu.

    Parameters:
    -----------

    permu : ndarray
      A one dimensional array representing a permutation.

    """

    num_algorithms = len(permu)
    for i in range(num_algorithms):
      idx = np.random.randint(i, num_algorithms)
      permu[i], permu[idx] = permu[idx], permu[i]
    return permu

  def calculate_top_ranking_probs(self, permus):
    """ Calculate the probability that each algorithm to be the best algorithm.

    Parameters:
    ------------

    permus : ndarray 
      Array of num_permus rows representing a permutation and num_algorithms columns 
      representing  the ranking of each algorithm.

    Returns
    --------

    probs : ndarray
      Array of size (num_posterior_samples, num_algorithms) where the 
      first dimension refers to the number of posterior samples and the second dimension
      represent each algorithm. For example, probs[10, 0] is the probability that the algorithm with
      the index 0 is the best algorithm according to the posterior sample with index 10.

    """

    samples = self.sample_posterior(permus)
    num_permus, num_algorithms = permus.shape
    num_posterior_samples = len(samples)
    probs = np.zeros((num_posterior_samples, num_algorithms))

    for i, sample in enumerate(samples):
      for permu in itertools.permutations(list(range(num_algorithms))):
        permu = np.array(permu)
        probs[i, permu[0]] += self.calculate_permu_prob(permu, sample)

    return probs

  def calculate_better_than_probs(self, permus):
    """ Calculate the probability that each algorithm to outperforms other
    algorithms. The function computes such probability for each possible
    pair of algorithms.

    Parameters:
    ------------

    permus : ndarray 
      Array of num_permus rows representing a permutation and num_algorithms columns 
      representing  the ranking of each algorithm.

    Returns
    --------

    probs : ndarray
      Array of size (num_posterior_samples, num_algorithms, num_algorithms) where the 
      first dimension refers to the number of posterior samples, the second dimension
      represent each algorithm and the third dimension represents the algorithm which is
      outperformed. For example, probs[10, 0, 2] is the probability that the algorithm with
      the index 0 outperforms the algorithm with the index 2 according to the posterior 
      sample with index 10.

    """

    samples = self.sample_posterior(permus)
    num_permus, num_algorithms = permus.shape
    num_posterior_samples = len(samples)
    probs = np.zeros((num_posterior_samples, num_algorithms, num_algorithms))

    for i, sample in enumerate(samples):
      for permu in itertools.permutations(list(range(num_algorithms))):
        permu = np.array(permu)
        prob = self.calculate_permu_prob(permu, sample)

        for j in range(num_algorithms):
          for k in range(j + 1, num_algorithms):
            probs[i, permu[j], permu[k]] += prob

    return probs

  def calculate_top_k_probs(self, permus):
    """ Calculate the probability that each algorithm is in the top-k
    ranking. The function computes such probability for each possible
    value of k for 1 < k < num_algorithms.

    Parameters:
    ------------

    permus : ndarray 
      Array of num_permus rows representing a permutation and num_algorithms columns 
      representing  the ranking of each algorithm.

    Returns
    --------

    probs : ndarray
      Array of size (num_posterior_samples, num_algorithms, num_algorithms) where the 
      first dimension refers to the number of posterior samples, the second dimension
      represent each algorithm and the third dimension represents the value of k being
      calculated. For example, probs[10, 0, 2] is the probability to be in the top-3 ranking 
      of the algorithm with index 0 according to the posterior sample with index 10.

    """

    samples = self.sample_posterior(permus)
    num_permus, num_algorithms = permus.shape
    num_posterior_samples = len(samples)
    probs = np.zeros((num_posterior_samples, num_algorithms, num_algorithms))

    for i, sample in enumerate(samples):
      for permu in itertools.permutations(list(range(num_algorithms))):
        permu = np.array(permu)
        prob = self.calculate_permu_prob(permu, sample)

        for j in range(num_algorithms):
          for k in range(j, num_algorithms):
            probs[i, permu[j], k] += prob

    return probs
